"""
Main Gateway class which manages all user modules.
"""
from typing import Callable, List, Dict, Union
import contextlib
import json
import threading
import sys
import signal
import logging
import requests

from .typing import IModule, Numeric, ModuleId
from .modules import init as init_base_modules
from .UuAppClient import UuAppClient
from .decorators import on, _registered_methods, _filter_registered_methods
from .utils import RepeatTimer

logger = logging.getLogger(__name__)
loggerc = logging.getLogger(__name__ + '.Config')

class ModuleTimer:
    """ Store :class:`~uun_iot.modules.Module.IModule` instance together with its tick event timer(s).

        ``f`` and ``period`` are of type `function` and `float` respectively:

            - ``f``: function to be called on each timer hit (module's method)
            - ``period``: period of the timer in seconds,

        or ``f`` and ``period`` can be either of type `dict` and `dict` respectively,
        indexed by the **same** keys, the keys correspond to timer IDs:

            - ``f``: values of the dictionary are the ``module_instance``'s method
            - ``period``: values are timer periods in seconds.

        Args:
            module_instance: an instantiated :class:`~uun_iot.modules.Module.IModule`
                (for which the timer is created)
            f: method of ``module_instance`` object to be repeatedly called on `tick` event
            period: the `tick` period in seconds

        Raises:
            ValueError: argument type mismatch for `f` and `period`, see above.
    """
    #: ID of passed module
    id: ModuleId
    m_instance: IModule

    #: does timer have multiple subtimers?
    multi: bool
    timer: Union[RepeatTimer, Dict[str, RepeatTimer]]

    def __init__(self,
            module_instance: IModule,
            f: Union[Callable, Dict[str, Callable]],
            period: Union[Numeric, Dict[str, Numeric]]
        ):
        self.m_instance = module_instance
        self.id = module_instance.id

        # timers are now instanced as a registered function (from callback) and
        #   the module instance is passed to it as the `self` argument

        if callable(f) and isinstance(period, (int, float)):
            # single (unnamed) timer
            self.multi = False
            self.timer = RepeatTimer(period, f,
                    kwargs={'self': self.m_instance}
                )
            logger.debug("assigned %ss timer to `%s`", period, self.id)

        elif isinstance(f, dict) and isinstance(period, dict):
            # multiple timers
            self.multi = True
            self.timer = {}
            for timer_id, f_single in f.items():
                try:
                    period_single = period[timer_id]
                except KeyError:
                    logger.error(
                        "Could not find timer `%s:%s` in configuration "
                        "(this module has multiple timers).",
                        self.id, timer_id
                    )
                    continue
                self.timer[timer_id] = RepeatTimer(period_single, f_single,
                        kwargs={'self': self.m_instance}
                    )
                logger.debug("assigned %ss timer to `%s:%s`", period_single, self.id, timer_id)
        else:
            raise ValueError("Argument type mismatch. Arguments `f` and `period` have to "
                    "be of types Callable and float, or Dict[str, Callable] and Dict[str, float]."
                    f"They are `{type(f)}` and `{type(period)}` instead.")


class Gateway(contextlib.AbstractContextManager):
    """Main Gateway class which manages all user modules.

    Gateway is responsible for managing all modules and also manages
    event dispatch and configuration.

    This class is a `ContextManager`. On enter, the :meth:`.start` method is called,
    on leave, the :meth:`.stop` method is called.

    Warning:
        When an exception is thrown inside the ``module_init`` function (usually calling class constructors), it is not supressed and leads to classical exception behaviour (ie. program termination). 

        When an exception is thrown inside a :meth:`~uun_iot.decorators.on` event hook, the exception is catched and printed via logging.

    Examples:

        .. code-block:: python

            from uun_iot import Gateway:
            config_file="config.json"
            with Gateway(config_file) as g:
                while True:
                    time.sleep(1)

        In this example, the :class:`Gateway` is initialized using the
        configuration file in a `with` block. Upon entering, the gateway is
        started. When an exeption occurs, the gateway is gracefully stopped
        using its :meth:`.stop` method -- this allows the `Gateway` to do
        cleanup and inform user modules (here are none) that they should exit,
        too.


    Args:
        config_file: path to configuration JSON file. See :class:`Config` for
            information about format of the file.
        module_init: a function ``(config_dict, uuapp_client) -> List[IModule]``
            or ``(config_dict, uuapp_client) -> IModule`` for single module.
            Here, ``config_dict`` is the dictionary created from JSON ``config_file``
            and ``uuapp_client`` is an instance of :class:`~uun_iot.UuAppClient.UuAppClient`.
            The function is responsible for initializing user modules.
    """

    #: This Event is set when the Gateway :meth:`start`-s and cleared when :meth:`stop`-s.
    runev: threading.Event
    #: This Event is set when the Gateway :meth:`stop`-s and cleared when :meth:`start`-s.
    stopev: threading.Event

    #:class:`Config` instance managed by Gateway
    _config_obj: "Config"
    #: dictionary storing the gateway's configuration
    #: see :attr:`Config.config`
    config: dict

    uuapp_client: UuAppClient

    #: list of managed user modules
    _modules: Dict[ModuleId, IModule]
    #: list of tick event timers
    _timers: List[ModuleTimer]

    def __init__(
            self,
            config_file: str,
            module_init: Union[
                Callable[[dict, UuAppClient], List[IModule]],
                Callable[[dict, UuAppClient], IModule]
                ]=None
        ):

        logger.debug("Starting gateway module system.")

        self.runev = threading.Event()
        self.stopev = threading.Event()

        self._config_obj = Config(self, config_file)
        self.config = self._config_obj.config

        self.uuapp_client = UuAppClient(self.config)
        modules = init_base_modules(self.config, self.uuapp_client)

        # additional modules specific to application
        if module_init:
            mi = module_init(self.config, self.uuapp_client)
            if isinstance(mi, list):
                modules += mi
            else:
                # a single module
                modules += [mi]

        # add Config as an equivalent of module
        # see Config class for further details
        modules.append(self._config_obj)

        self._modules={}
        for m in modules:
            if not hasattr(m, "id") or m.id is None:
                logger.error("A user module does not have attribute 'id' or it "
                             "is set to None! Check that you properly define "
                             "'id' in all modules.")
                exit(1)
            else:
                self._modules[m.id] = m

        # clear some remnants of mass importing (for ex. when testing)
        _filter_registered_methods(list(self._modules.keys()))

        # register `on("update")` callbacks with the Config module which is responsible for updates
        self._config_obj.register_callbacks(self._modules)

        # register timers
        self._init_timers()

        logger.debug("module system initilized")

    def __enter__(self):
        """ Call :meth:`.start`. """
        self.start()
        return self

    def __exit__(self, *exc_info):
        """ Call :meth:`.stop`. """
        self.stop()

    def signal_handler(
            self,
            sig: signal.Signals,
            frame,
            additional_cleanup: Callable[["Gateway"], None]=None
        ):
        """ Handler for predefined signals.

        The following signals are supported:

        =================   ===========
        functionality no.   description
        =================   ===========
        1                   :meth:`.stop` Gateway, ``del`` ete all modules, run ``additional_cleanup`` and exit
        =================   ===========

        =========== =============
        signal      functionality
        =========== =============
        ``SIGINT``  1
        ``SIGTERM`` 1
        ``SIGUSR1`` 1 and exit with error code ``1``
        ``SIGUSR2`` 1
        =========== =============

        The signals need to be explicitly registered with this method being the
        associated handler. Register signals as:

        .. code-block:: python

            import signal
            from uun_iot import Gateway
            with Gateway(...) as g:
                signal.signal(signal.SIGTERM, g.signal_handler)
                ...

        If you want to specify ``additional_cleanup``, define a partial
        function from this method with first two arguments left empty.

        Args:
            sig: caught signal
            frame: exception frame
            additional_cleanup: optional function to be run to cleanup the
                `Gateway`. The function takes the `Gateway` instance as an
                argument.
        """
        logger.debug("Received signal `%s`", signal.Signals(sig).name)
        if sig in [signal.SIGINT, signal.SIGTERM, signal.SIGUSR1, signal.SIGUSR2]:
            # systemd issues SIGTERM to stop a program
            # user defined signal SIGUSR1 -- is issued from some Module to stop gateway and exit
            self.stop()

            # explicitly delete a module before exiting to give it a chance to execute its __del__
            for (mid, module) in self._modules.items():
                del module

            if additional_cleanup is not None:
                additional_cleanup(self)

            ec = sig in [signal.SIGUSR1]
            # SIGUSR1 -- exit with error
            # SIGUSR2 -- exit without error
            sys.exit(int(ec))

    def start(self):
        """Invoke all ``@on("start")`` functions and start all timers.

        This is no-op when the gateway is already started.
        """
        if not self.runev.is_set():
            self.runev.set()
            self.stopev.clear()
            self._invoke_on_start()
            self._start_timers()

    def stop(self):
        """
        Invoke all ``@on("stop")`` functions and peacefully stop (wait to finish)
        all timers and associated ``@on("tick")`` handlers.

        This is no-op when the gateway is already stopped.
        """

        # this also stops on_start function (oneshots) polling for the event
        if self.runev.is_set():
            self.runev.clear()
            self.stopev.set()
            self._invoke_on_stop()
            self._stop_timers()

    def _init_timers(self):
        """ Initialize tick event timers with interval(s) from configuration. """
        self._timers = {}
        for (mid, module) in self._modules.items():
            try:
                interval = self.config["gateway"]["moduleTimers"][mid]
            except KeyError:
                # module not present in configuration
                logger.info("The initialized module `%s` has no corresponding timer configuration "
                "entry. Did you forget to make a configuration entry for its timer?", mid)

                logger.debug("Disabling timer for module `%s`", mid)
                continue

            callback = _registered_methods['on_tick'][mid]
            self._timers[mid] = ModuleTimer(module, callback, interval)

    def _start_timers(self):
        """
        Start tick event timers for modules.
        """
        for mid, m_timer in self._timers.items(): #module_id, module_timer object
            if m_timer.multi:
                for tid, timer in m_timer.timer.items():
                    timer.start()
                    logger.debug("Timer `%s:%s` started.", mid, tid)
            else:
                m_timer.timer.start()
                logger.debug("Timer `%s` started.", mid)

    def _stop_timers(self):
        """
        Stop timers for modules.
        """
        for mid, m_timer in self._timers.items():
            if m_timer.multi:
                for tid, timer in m_timer.timer.items():
                    timer.stop()
            else:
                m_timer.timer.stop()

    def _invoke_on_start(self):
        """ Start each of ``@on("start")`` methods in its own thread. """
        for mid, func in _registered_methods["on_start"].items():
            # do not store the thread to kill it later, rely solely on the func
            #   to terminate itself when runev Event is unset
            logger.debug("Calling @on('start') function `%s`", func)
            threading.Thread(
                    target=func,
                    args=(self._modules[mid], (self.runev, self.stopev)) # self, (runev, stopev)
                ).start()
    def _invoke_on_stop(self):
        """ Start each of ``@on('stop')`` methods in its own thread. """
        for mid, func in _registered_methods["on_stop"].items():
            logger.debug("Calling ``@on('stop')`` function `%s`", func)
            threading.Thread(
                    target=func,
                    args=(self._modules[mid],) # self
                ).start()

# Class Config is in the same file as Gateway due to intimate connection of the two.
# Split would result in cyclic import or unneccesarry generalization in rest of
# the code only for this one case.
class Config(IModule):
    """ Configuration module.

    The preffered styling of the JSON keys is ``camelCase``, ie. first letter
    lowercase, letters after spaces uppercase and finally, the spaces removed.

    The JSON configuration file has the following example structure:

    .. code-block:: json
        :force:

        {
            "gateway": {

                "moduleTimers": {
                    "customModule1": {
                        "receive": 120,
                        "send": 400
                    },
                    "customModule2": 60
                },

                "moduleBackupStorage": {
                    "customModule1": "backup/customModule1.json",
                    "customModule2": {
                        "path": "backup/customModule2.json",
                        "limit": 50
                    }
                },

                "customModule2": {
                    "option1": "value1"
                }
            },

            "uuApp": { ... },
            "uuThing": { ... },
            "oidcGrantToken": { ... }
        }

    On the other hand, a **minimal** configuration file is an empty JSON file:

    .. code-block:: json

        {}

    Meaningful keys and subkeys are:

        - ``gateway``: `optional`. Main configuration for this IoT application

            - ``moduleTimers``: `optional`. Core functionality, dictionary with
              periods (in seconds) for ``@on("tick")`` events for corresponding
              modules. The module IDs are keys, the periods are values.

                - Multiple timers can be
                  specified by introducing a subobject with the timer ID and the
                  timer period. The timer ID corresponds to the event defined in
                  Python code as ``@on("tick", "timerId")``.

            - ``moduleBackupStorage``: `optional`. Applies to modules based on
              :class:`~uun_iot.modules.Module.Module`. The format is module ID
              as key and file path as value. This key is used to specify
              location to which should unsent data from the module be saved,
              see :class:`~uun_iot.modules.Module.Module` for more information.

                - You can specify additional information, such that the storage
                  should be limited in size. For this, specify the size of the
                  storage in number of entries in ``limit`` and add the
                  original file path in the ``path`` key.

            - ``<moduleId>``: `optional`. A default place for the configuration specific to
              the module with ID ``moduleId``. The structure is arbitrary and
              depends on your needs.

        - keys for :class:`uun_iot.UuAppClient`. `Optional`. See documentation there for
          more information. If you want to use secure communication with uuApp,
          specify the details in keys

            - ``uuApp``
            - ``uuThing``
            - ``oidcGrantToken``



    Args:
        gateway_instance: :class:`Gateway` instance
        config_file: path to configuration JSON file
    """

    id = 'config'

    def __init__(self, gateway_instance: Gateway, config_file: str):
        self.g = gateway_instance
        self._config_file = config_file
        with open(config_file, "r") as f:
            self.config = json.load(f)

        if "gateway" not in self.config:
            self.config["gateway"] = {}

    def _uucmd(self):
        # create uucmd function, separate from other cmd due to nature of Config
        uucmd = self.config["uuApp"]['uuCmdList']['gatewayGetConfiguration']
        try:
            r = self.g.uuapp_client.get_request(uucmd)
            r.raise_for_status()
            new_config = r.json()
            # this is not an actual configuration key, only a remnant of server communication
            del new_config["uuAppErrorMap"]
        except requests.exceptions.HTTPError as e:
            loggerc.warning("Could not fetch new data (error code %s)", r.status_code)
            loggerc.debug("Received response: %s", r.text)
            return False
        except requests.exceptions.RequestException as e:
            loggerc.info("Error in server communication: %s", e)
            return False
        except json.decoder.JSONDecodeError:
            loggerc.error("Received invalid configuration JSON from server")
            loggerc.debug("Invalid response: %s", r.text)
            return False
        except KeyError:
            pass

        if new_config == {}:
            loggerc.error("Received invalid JSON configuration from server")
            loggerc.debug(r.text)
            return False

        return new_config

    def register_callbacks(self, modules):
        """ Register optional update event callbacks with :class:`Config`. """
        self._callbacks = {}
        for (mid, module) in modules.items():
            if mid in _registered_methods['on_update']:
                # store along with module instance
                self._callbacks[mid] = (_registered_methods['on_update'][mid], module)
                loggerc.debug("Registered update callback for `%s`.", mid)

    @on("tick")
    def on_tick(self):
        """
        Gets new configuration from the uuApp, validates it, restarts timers
        (if needed based on the new configuration) and saves the new
        configuration locally. It also notifies modules (via their
        ``@on("update")``) about configuration update.
        It is triggered by a tick event.
        """

        new_config = self._uucmd()

        if not new_config:
            # did not get a valid configuration
            return

        if new_config != self.config["gateway"]:

            loggerc.debug("Received new configuration from server: %s", new_config)

            # for loop: restart timers if needed
            for id, mt in self.g._timers.items():
                if id not in new_config["moduleTimers"]:
                    continue
                    
                # new timer values are presented
                if self.config["gateway"]["moduleTimers"][id] != new_config["moduleTimers"][id]:

                    if id == "config":
                        # TODO: needs testing if utils.Timer correctly handles setting period without stopping first
                        mt.timer.period = new_config["moduleTimers"][id]
                        loggerc.info(f"experimentally updated `{id}`'s timer to {new_config['moduleTimers'][id]} s")
                        continue

                    if mt.multi:
                        # multiple timers for one module
                        if self.config["gateway"]["moduleTimers"][id].keys() != new_config["moduleTimers"][id].keys():
                            loggerc.error("Update cannot introduce new timer_id for a module, only update timer values.")
                            continue

                        for timer_id, new_interval in new_config["moduleTimers"][id].items():
                            if new_interval != self.config["gateway"]["moduleTimers"][id][timer_id]:
                                mt.timer[timer_id].stop()
                                mt.timer[timer_id].period = new_config["moduleTimers"][id][timer_id]
                                mt.timer[timer_id].runonstart = False
                                mt.timer[timer_id].start()
                                loggerc.info(f"updated `{id}:{timer_id}` timer to {new_config['moduleTimers'][id][timer_id]} s")
                    else:
                        # single timer
                        mt.timer.stop()
                        mt.timer.period = new_config["moduleTimers"][id]
                        mt.timer.runonstart = False
                        mt.timer.start()
                        loggerc.info(f"updated `{id}`'s timer to {new_config['moduleTimers'][id]} s")

            # update config, only gateway key, rest is unchanged

            # PASSIVE CONFIGURATION UPDATE
            # propagated passively to each module
            # modules contain _config in form of config["gateway"] assigned at module's init
            #   (other keys are not needed and only cause more typing)
            
            # DOES NOT WORK:
            # cannot do `self.config["gateway"] = new_config` as it would not change the config["gateway"] dict present in each module:
            # self.config["gateway"]           --> { 'module_1' --> module_1_data_old, ... }
            # self.config["gateway"] --(update)--> { 'module_1' --> module_1_data_new, ... }
            # (some_module)._config            --> { 'module_1' --> module_1_data_old, ... } as
            #   it was never reassigned inside the module
            
            # DOES WORK:
            # update every key(pointer) in config["gateway"] separately; module's _config still points to the same config["gateway"]
            #   but every key inside config["gateway"] is correctly updated
            # self.config["gateway"] --> { 'module_1'           --> module_1_data_old, ... }
            # self.config["gateway"] --> { 'module_1' --(update)--> module_1_data_new, ... }
            # (some_module)._config  --> { 'module_1'           --> module_1_data_new, ... } as 
            #   the module's config and self.config["gateway"] point to the same (updated object)
            
            # new update cannot introduce new keys to ["gateway"] (which is to be expected -- cannot create objects at run-time without any defining code)
            updated_values = False
            for key in self.config["gateway"]:

                if key not in new_config:
                    loggerc.warning("Update does not contain a configuration key `%s`. Update is not deleting this key from current configuration but it will be deleted on next run of application.", key)
                    continue

                if self.config["gateway"][key] != new_config[key]:
                    updated_values = True
                    loggerc.debug("changed key `%s` from %s to %s", key, self.config["gateway"][key],  new_config[key])
                    self.config["gateway"][key] = new_config[key]

            if not updated_values: 
                loggerc.debug("New configuration introduced only new keys not present on old config, discarding update.")
                return

            # allow module callbacks to alter the configuration before saving to disk
            self._invoke_update_callback()

            # save updated config to file in case of power outage/app restart
            with open(self._config_file, 'w') as f:
                f.write(json.dumps(self.config, indent=4)) # pretty print in case of manual editing

            loggerc.info("Updated configuration.")

    def _invoke_update_callback(self):
        """ Notify all modules (which support it) about configuration change. """
        for id, (callback, m_instance) in self._callbacks.items():
            threading.Thread(
                    target=callback,
                    args=(m_instance,)
                ).start()




__version__ = "0.9.1"

from .typing import IModule
from .decorators import on
from .UuAppClient import UuCmdSession
from .Gateway import Gateway
from .modules.Module import Module
from .modules.Module import ConfigScopeEnum



""" Types and interfaces frequently used in uun-iot library and apps. """

from abc import ABC, abstractmethod
from typing import Union

ModuleId = str
Numeric = Union[float, int]

class IModule(ABC):
    """ Module interface. All user modules must inherit from this interface. """
    id: str

    @property
    @abstractmethod
    def id(self) -> str:
        pass



from typing import Dict, List
from threading import Lock, Event
import signal
import os
import datetime
import logging

from uun_iot import on
from uun_iot.typing import ModuleId
from uun_iot.diagnostic import GeneralDiagnosticEvent, DiagnosticEvent, GatewayDiagnosticEvent
from .Module import Module

logger = logging.getLogger(__name__)

class BaseHealthCheck(Module):
    """ Base class for creating runtime consistency checks and debug diagnostics.

    HealthCheck modules can log incoming events and act correspondingly, ie. a
    HealthCheck module can restart application if the app is not working
    correctly and log more debugging information for further analysis.

    This can be useful for detecting anomalous cases when the app is not
    working correctly due to an unknown bug, typically when the bug occurs at
    weird edge cases in long running apps. At the same time, it allows to
    recover from the case (by restarting) and to collect diagnostic info.

    HealthCheck can monitor multiple modules at the same time. This is done
    using ``@on("tick", "moduleId")`` timers.

    The monitored modules need to explicitly send information to the
    HealthCheck module as this cannot always be automatized.

    Event informations are submitted by monitored modules via :meth:`.notify` method
    in a form of :class:`uun_iot.diagnostic.GeneralDiagnosticEvents`.

    BaseHealthCheck only provides infrastructure for storing these events and
    it is upto app-specific HealthCheck to implement neccessary logic for
    determining malfunctioning app states. Although, some helper functions for
    usual cases are already present in BaseHealthCheck
    (:meth:`._act_if_data_not_sent`).

    BaseHealthCheck can register basic @on methods automatically. This is
    useful, as most of derived HealthCheck modules will need
    `save_all_diagnostics` registered as `@on(stop)` and `_init_module_config`
    registered as `@on(update)` for updating module-specific leeways etc.

    Sample gconfig:

    .. code-block:: json

         {
            "moduleTimers": {
                "healthCheck": {
                    "weatherConditions": 300
                }
            }

            "healthCheck": {
                "diagnosticPath": "diagnostics/",
                "modules": {
                    "weatherConditions": {
                        "leeway": 120
                    }
                }
            }
        }

    Leeways should be sufficiently large (larger than watched module's timers),
    so that the module can do its function well. Only module-level configuration
    is dynamically updateable: leeways, 

    Args:
        gconfig (dict): gateway configuration
        on_stop (bool): default True, assign ``@on("stop")`` decorator
            to :meth:`._save_all_diagnostics`
        on_update (bool): default True, assign `@on("update")` decorator
            to :meth:`._init_module_config`
    """

    _start_date: datetime.datetime
    _leeways: Dict[ModuleId, datetime.timedelta]
    _diagnostic_active: Dict[ModuleId, Event]
    _diagnostic_log: Dict[ModuleId, List]
    _event_dates: Dict[ModuleId, Dict[GeneralDiagnosticEvent, datetime.datetime]]
    _diagnostic_lock: Dict[ModuleId, Lock]
    _diagnostics_save_path: str

    def __init__(self, gconfig: dict, on_stop: bool=True, on_update: bool=True):
        """
        """
        super().__init__(config=gconfig)

        self._diagnostics_save_path = self._c("diagnosticsPath").rstrip("/")

        self._diagnostic_lock = {
                mid: Lock() for mid in self._c("modules") 
            }

        self._diagnostic_active = {
                mid: Event() for mid in self._c("modules")
            }

        self._diagnostic_log = {
                mid: [] for mid in self._c("modules")
            }

        self._event_dates = {
                mid: {} for mid in self._c("modules") 
            }
        # global information about Gateway, Event can be any of GatewayDiagnosticEvent
        self._event_dates["gateway"] = {}
        self.notify("gateway", GatewayDiagnosticEvent.GATEWAY_START)
        # alias for quick access, start time of gateway
        self._start_date = self._event_dates["gateway"][GatewayDiagnosticEvent.GATEWAY_START]

        self._init_module_config()

        self._stop_diagnostic = self.save_diagnostic

        if on_stop:
            on("stop")(self.save_all_diagnostics)
        if on_update:
            on("update")(self._init_module_config)

    def _init_module_config(self):
        """ Initialize module-specific configuration of monitored modules from configuration.
        Ie. everything under `modules.<moduleId>.*` configuration.
        Currently only leeways."""
        self._leeways = {
                mid: datetime.timedelta(seconds=self._c(f"modules/{mid}/leeway")) for mid in self._c("modules")
            }

    def notifier(self, mid: ModuleId):
        """ Return a :meth:`.notify` method with prefilled ``mid`` argument.

        Args:
            mid: module ID
        """
        return lambda *args, **kwargs: self.notify(mid, *args, **kwargs)

    def notify(self, mid: ModuleId, event: GeneralDiagnosticEvent, *args, **kwargs):
        """Notify HealthCheck about watched modules' events.

        Optionally, add more custom data about the event in form of additional
        ``*args`` or ``**kwargs``.

        Args:
            mid: watched module ID
            event: GeneralDiagnosticEvent representing what action was done
            \*args: other positional arguments for logging purposes
            \*\*kwargs: other keyword arguments for logging purposes
        """
        now = datetime.datetime.now()
        if not mid in self._event_dates:
            raise ValueError("Module is not listed in HealthCheck's configuration. Add it there first.")
        if not isinstance(event, GeneralDiagnosticEvent):
            raise ValueError("Event should inherit from uun_iot.diagnostic.GeneralDiagnosticEvents.")

        self._event_dates[mid][event] = now

        if mid in self._diagnostic_active and self._diagnostic_active[mid].is_set():
            entry = (now, event.name, (args, kwargs))
            with self._diagnostic_lock[mid]:
                self._diagnostic_log[mid].append(entry)

    def start_diagnostic(self, mid):
        """ Start logging incoming events for module `mid`. """
        self._event_dates[mid][DiagnosticEvent.DIAGNOSTICS_START] = datetime.datetime.now()
        self._diagnostic_active[mid].set()

    def save_all_diagnostics(self):
        """ Stop all diagnostics and save them to files. """
        for mid in self._c("modules"):
            self.save_diagnostic(mid)

    def save_diagnostic(self, mid):
        """ Stop logging incoming events for `mid` and save them to file. """
        with self._diagnostic_lock[mid]:
            if not self._diagnostic_active[mid].is_set():
                return

            try:
                logger.info(f"Saving diagnostic of module {mid}.")

                dname = self._diagnostics_save_path
                mdname = f"{dname}/{mid}"
                os.makedirs(dname, exist_ok=True)
                os.makedirs(mdname, exist_ok=True)

                stop_d = datetime.datetime.now()
                self._event_dates[mid][DiagnosticEvent.DIAGNOSTICS_STOP] = stop_d

                # weatherConditions_2021-09-26_19-02-56UTC
                # diagnostic file creation
                start_d = self._event_dates[mid][DiagnosticEvent.DIAGNOSTICS_STOP].strftime("%Y-%m-%d_%H-%M-%S%Z")
                extension = "log"
                fname = f"{mid}_diagnostic_{start_d}"

                # find a nonused fname
                i = 0
                while os.path.exists(f"{mdname}/{fname}.{extension}"):
                    i+=1
                    fname = fname.split(".")[0] + f".{i}"
                fname = f"{mdname}/{fname}.{extension}"
                
                dlog = self._diagnostic_log[mid]
                with open(fname, "w") as f:
                    infile_date_format = "%Y-%m-%d %H.%M.%S.%f %Z"
                    start_d = self._event_dates[mid][DiagnosticEvent.DIAGNOSTICS_START].strftime(infile_date_format)
                    stop_d = stop_d.strftime(infile_date_format)
                    f.write(f"Diagnostics of module {mid}\nStarts at: {start_d}\nEnds at:   {stop_d}\n\n")

                    if dlog == []:
                        f.write(f"No data were collected (HealthCheck notifiers were not reached in target code)\n")
                    else:
                        for d in dlog:
                            date = d[0].strftime(infile_date_format)
                            event = d[1]
                            arg = d[2]
                            if arg is None or arg == ((), {}):
                                f.write(f"{date}: {event}\n")
                            else:
                                args = arg[0]
                                kwargs = arg[1]
                                sa = ", ".join(list(map(str, args)))
                                sk = ", ".join([f"{k}={v}" for k,v in kwargs.items()])
                                if sa == "":
                                    f.write(f"{date}: {event} - {sk}\n")
                                else:
                                    f.write(f"{date}: {event} - {sa}; {sk}\n")


                self._diagnostic_log[mid] = []
                self._diagnostic_active[mid].clear()
            except Exception as e:
                print("_diagnostic_log[mid]: ", self._diagnostic_log[mid])
                print("_diagnostic_active[mid]", self._diagnostic_active[mid])
                print("_diagnostic_lock[mid]", self._diagnostic_lock[mid])

                self._diagnostic_log[mid] = []
                self._diagnostic_active[mid].clear()

                logger.error("Error occured while saving diagnostics:")
                raise e

    def _exit_app(self):
        # signal SIGUSR1 to itself
        # Gateway will handle it and gracefully exit the whole app
        self.notify("gateway", GatewayDiagnosticEvent.GATEWAY_SIGNAL, "SIGUSR1")
        os.kill(os.getpid(), signal.SIGUSR1)

    def _act_if_data_not_sent(self, mid: ModuleId, send_leeway: datetime.timedelta, now: datetime.datetime):
        """
        Check if module (usually its uucmd send handler) sent data in time
        interval (now-send_leeway[mid], now).

        If yes, save any current running diagnostic for `mid`. If not, run
        diagnostic - or - shutdown the application, if diagnostic is already
        running.

        Note:
            Even if the module has no data to send, this will flag it as
            malfunctioning and restart the app.

        Data send event is considered to be DiagnosticEvent.DATA_SEND_ATTEMPTED.
        No internet server communication is needed, only local information.

        .. code-block::

                               <--------> send_leeway
            -*-----------------*----------------------| now
             ^ attempted sends ^

        Args:
            mid: module id
            send_leeway (datetime.timedelta): specified time window (relative
                to now) to check for sent data
            now (datetime.datetime): current time
        """
        start_time = self._event_dates["gateway"][GatewayDiagnosticEvent.GATEWAY_START]

        # this handles the case when data were not yet sent (soon after start of gateway)
        dmin = datetime.datetime.min
        last_attempted_send = self._event_dates[mid].get(DiagnosticEvent.DATA_SEND_ATTEMPTED, dmin)

        if last_attempted_send == dmin:
            last_attempted_send = start_time

        if last_attempted_send + send_leeway < now:
            # Suspicious case:
            # module did not (even) tried sending any data during leeway.
            # gather diagnostic data about the situation and if this case
            # occurs once again, restart the application
            # (this should be solving the problem until the bug is fixed by
            # developer in new app version)
            if self._diagnostic_active[mid].is_set():
                logger.warning("Module `%s` again did not sent any data - exiting app.", mid)
                self._exit_app()
            else:
                logger.warning("HealthCheck detected an error in module `%s`: "
                    "no data are being sent.", mid)
                self.start_diagnostic(mid)
        else:
            # everything is ok now
            logger.debug("HealthCheck did not detect an error in module %s", mid)
            self._stop_diagnostic(mid)


        # ASSUMPTIONS:
        # uucmd handler signals <==> the handler tried
        # (successfully/unsuccessfully) sending data to server

        # OLD:
        # ASSUMPTIONS: uucmd handler signals offline <==> gateway is really offline
        # ie. no bugs in uucmd handler which would cause nothing to be sent,
        # even if the server is accessible
        # ALTERNATIVE approach: send only one signal (do not split into online/offline) in uucmd
        # this nested condition will reduce to only one if (if uucmd fired)
        # online & offline distinction is good for diagnostic only
        # ie. we only need to know IF uucmd handler fired during leeway,
        #     no need to know if it was online or offline

        # note: order of online/offline is irrelevant, what matters is if uucmd fired
        # if lonline + self.leeway < now:
        #     # suspicious case
        #     # this might mean that there has been no internet during leeway
        #     # OR that weathercondition module (WCM) is not sending data (due to some unknown bug)

        #     if loffline + self.leeway < now:
        #         # the WCM has attempted to send data but did not succeed
        #         # (offline state is signalled <=> the WCM uucmd tried to send data)
        #         # ==> there is no internet, nothing to do about it



from datetime import datetime
import logging
import requests
import psutil
from typing import List, Dict, Callable

from uun_iot.utils import get_iso_timestamp
from uun_iot.decorators import on
from uun_iot.typing import IModule

logger = logging.getLogger(__name__)

class Heartbeat(IModule):
    """
    Ping server in short periods to let server know if the gateway is online and
    send a little info about the gateway.

    Args:
        uucmd: function ``(dto_in) -> requests.Response``, the function takes
            an argument with data to be sent to the heartbeat uuCmd endpoint. It
            returns the reponse formed using :class:`requests.Response`.
    """

    id="heartbeat"

    def __init__(self, uucmd: Callable[[Dict], requests.Response]):
        self.online = False
        self._uucmd = uucmd

    @on("tick")
    def on_tick(self):
        """Determine online status and send a little information about gateway to uuApp.

        The online status is determined using ``self._uucmd`` uuCmd. The information sent includes

            - boot timestamp (in ISO timestamp)
            - CPU usage in percent
            - RAM usage in percent

        Executed on each timer tick.
        """
        boot_timestamp = datetime.fromtimestamp(psutil.boot_time())
        dto_in = {
            "bootTimestamp": get_iso_timestamp(boot_timestamp),
            "cpuUsage": psutil.cpu_percent(),
            "ramUsage": psutil.virtual_memory().percent
        }

        response = None
        try:
            response = self._uucmd(dto_in)
        except requests.exceptions.ConnectionError as e:
            logger.warning("Error connecting to server: %s", e)

        if response:
            if response.status_code >= 200 and response.status_code < 300:
                if not self.online:
                    logger.info("online")
                    self.online = True
        else:
            if self.online:
                logger.warning("offline")
                self.online = False



from enum import IntEnum, auto

class GeneralDiagnosticEvent(IntEnum):
    """ Base class for diagnostic events specified by uun-iot library or user
    applications. Diagnostic events are used to notify HealthCheck about some
    important events in via `HealthCheck.notify`. If user application needs
    more diagnostic events, inherit from this class. """
    pass

class DiagnosticEvent(GeneralDiagnosticEvent):
    """ General diagnostic events. """

    DIAGNOSTICS_START = auto()
    """ Start of diagnostics. """

    DIAGNOSTICS_STOP = auto()
    """ Stop of diagnostics. """

    TIMER_CALL = auto()
    """ Registered @on(tick) method is called.
    Specify timer ID in custom argument for better debugging purposes. """

    ON_START_CALL = auto()
    """ Registered @on(start) method is called. """

    ON_STOP_CALL = auto()
    """ Registered @on(stop) method is called. """

    ON_UPDATE_CALL = auto()
    """ Registered @on(update) method is called. """


    DATA_RECEIVED = auto()
    """ Module received valid data. """

    DATA_SEND_IMMINENT = auto()
    """ Module is going to send data. Use directly in front of `self._send_storage()` call. """

    DATA_SEND_OK = auto()
    """ Module has sent all data successfully. """

    DATA_SEND_PARTIAL = auto()
    """ Module has sent some data successfully, but some data were not sent. """

    DATA_SEND_FAIL = auto()
    """ Module could not send any data due to an error. """

    DATA_SEND_ATTEMPTED = auto()
    """ Module could not send any data due to an error. """


class GatewayDiagnosticEvent(GeneralDiagnosticEvent):
    """ Events tied directly to the Gateway instance. """

    """ Gateway was started. """
    GATEWAY_START=auto()

    """ Gateway was stopped. """
    GATEWAY_STOP=auto()

    """ Gateway received a signal.
    Specify signal type in custom argument for better debugging purposes."""
    GATEWAY_SIGNAL=auto()




"""
Event handler registration using ``@on`` decorator.
"""

from typing import List, Callable, Any, Dict, Union
from .typing import IModule
from .utils import module_id_from_str

#: dict: Global dictionary ``_registered_methods`` is a way to register
#:       modules' event handlers before inicialization of Gateway.
_registered_methods: Union[
        Dict[str, Dict],
        Dict[str, Dict[str, Callable]],
        Dict[str, Dict[str, Dict[str, Callable]]]
    ]
_registered_methods = {
    'on_tick': {},
    'on_update': {},
    'on_start': {},
    'on_stop': {},
}

# Exception for specifying unknown event in `on()` function
class UnsupportedEvent(Exception):
    """ Unknown gateway event. """

def _filter_registered_methods(module_names: List[str]):
    """Delete event handlers, which do not come from given list ``module_names``.

    The non-present event handlers are removed from :attr:`~uun_iot.decorators._registered_methods`.
    This is used internally to clear some remnants of mass importing (for ex. when testing).

    Args:
        module_names: list of module IDs. Any event handlers not belonging to
            these modules will be deleted from :attr:`~uun_iot.decorators._registered_methods`.
    """
    to_delete = []
    for event_types, event_handlers in _registered_methods.items():
        for module_name in event_handlers.keys():
            if module_name not in module_names:
                to_delete.append((event_types, module_name))

    for event_types, module_name in to_delete:
        del _registered_methods[event_types][module_name]

def _unbound_function(f: Callable) -> Callable:
    """
    If the function :func:`f` is bounded (ie. it is a method in some object instance,
    unbound it and return it. Otherwise, return original :func:`f`.
    """
    if hasattr(f, "__self__"):
        f = f.__func__
    return f

def _get_module_id_from_method(f: Callable) -> str:
    """
    Get corresponding class/module name of the passed method/function and get module ID
    using :func:`uun_iot.utils.module_id_from_str`. If the method
    is bound, return class name of the bound `self` object.

    Args:
        f: method belonging to class instance, alternatively a function belonging to Python module

    Returns:
        str: Module id to which the method belongs, alternatively a module id
        of Python module to which the function belongs
    """
    if hasattr(f, "__self__"):
        # f is a bound method (self argument is automatically prefilled)
        clsname = f.__self__.__class__.__name__
        if isinstance(f.__self__, IModule):
            return f.__self__.id
    else:
        clsname = f.__qualname__.split('.')[-2]
    mid = module_id_from_str(clsname) #clsname[0].lower() + clsname[1:]
    return mid

def on(*outer_args):
    """ Decorator for event handler registration.

    Synopsis: ``on(event[,id])``

    Supported events: ``tick``, ``update``, ``start``, ``stop``.

    ============= =======================================
        event                   description              
    ============= =======================================
     ``update``    gateway's configuration was updated   
     ``tick``      timer tick (configured in JSON)       
     ``start``     `Gateway` just started (using :meth:`~uun_iot.Gateway.Gateway.start`)
     ``stop``      `Gateway` is stopping (using :meth:`~uun_iot.Gateway.Gateway.stop`), end module's action as soon as possible
    ============= =======================================

        - ``tick`` event can take another argument to specify the timer id 
          (see ``gateway.moduleTimers`` key in configuration JSON), ie. ``on(fn)``,
          or  ``on(fn, "timerId")`` with :func:`fn` being the handler function.

    Handlers registered for the corresponding event with this decorator will be
    called on corresponding event by :class:`~uun_iot.Gateway.Gateway` or
    :class:`~uun_iot.Gateway.Config` objects. Passed arguments are different for each
    event and the methods have to take these arguments. Note: ``self`` denotes
    method's module instance, `origin` indicates which module initiates the event


    ============= ======================= ===================================
        event       handler synopsis             origin           
    ============= ======================= ===================================
     ``update``    ``handler(self)``        :class:`~uun_iot.Gateway.Config`   
     ``tick``      ``handler(self)``        :class:`~uun_iot.Gateway.Gateway`  
     ``start``     ``handler(self, evs)``   :class:`~uun_iot.Gateway.Gateway`   
     ``stop``      ``handler(self)``        :class:`~uun_iot.Gateway.Gateway`   
    ============= ======================= ===================================

    where ``evs = (g.runev, g.stopev)`` is a tuple of :class:`threading.Event`
    attributes :attr:`~uun_iot.Gateway.Gateway.runev` and
    :attr:`~uun_iot.Gateway.Gateway.stopev`. Here, ``g`` is the corresponding
    :class:`~uun_iot.Gateway.Gateway` instance.

    Note:

        In a typical use case, `@on` decorators are invoked on method/class definition,
        not at run-time. This can be seen on examples below.

    Examples:

        - ``timer`` event without ID

            - configuration:

                .. code-block:: json

                    {
                        "gateway": {
                            "moduleTimers": {
                                "timerModule": 1
                            }
                        }
                    }

            .. code-block:: python


                from uun_iot import on, Module
                class TimerModule(Module):
                    @on("tick")
                    def periodical(self):
                        print("Tick tock every 1 s.")

        - ``timer`` event with ID

            - configuration

                .. code-block:: json

                    {
                        "gateway": {
                            "moduleTimers": {
                                "sendReceive": {
                                    "send": 2,
                                    "get": 1
                                }
                            }
                        }
                    }

            .. code-block:: python

                class SendReceive(Module):
                    @on("tick", "get")
                    def get(self):
                        print(f"Retrieving data...")

                    @on("tick", "send")
                    def send(self):
                        print(f"Sending data...")

        - ``start`` event

            - configuration:

                .. code-block:: json

                    {
                        "gateway": {}
                    }

            .. code-block:: python

                class AdvancedDataMeasurement(Module):
                    @on("start")
                    def oneshot(self, evs):
                        runev, stopev = evs
                        while runev.is_set():
                            print("Polling for voltage reading from voltmeter...")
                            data = 53.8
                            if data > 50:
                                time.sleep(1)
                            else:
                                time.sleep(1.5)
                                print(data)

    Warning:

        *Dev note TODO.* Is method unbounding needed?
        Why is the decorated method being unbound here in the first place?

    Args:
        event (str): one of ``tick``, ``update``, ``start``, ``stop``
        id (str): optional, ID of the corresponding event, when more are
            specified in configuration JSON
    """
    eventarg = outer_args[0]
    event = "on_" + eventarg
    def wrapper(f):
        module_id = _get_module_id_from_method(f)
        f = _unbound_function(f)
        if event in _registered_methods.keys():
            if event == 'on_tick' and len(outer_args[1:]) == 1:
                _on_tick_id(f, outer_args[1])
            else:
                _registered_methods[event][module_id] = f
        else:
            raise UnsupportedEvent(f"Event type '{eventarg}' is not recognized. "
                    "Known events are:"
                    f"`{[k.replace('on_','') for k in _registered_methods]}`.")
        return f
    return wrapper

def _on_tick_id(f: Callable, timer_id: str):
    """
    Do not use directly! Use `@on(event)` instead. Registers a timer with specified id.

    If ``@on("tick")`` was already defined without ID, undefine the old handler.

    Args:
        f: Handler for the ``tick`` event.
        timer_id: ID of the timer as specified in configuration JSON

    """
    mid = _get_module_id_from_method(f)
    try:
        _registered_methods['on_tick'][mid][timer_id] = f
    except (TypeError, KeyError):
        # TypeError: storage was previously initialized with on("tick")
        #   - delete storage and save only this one
        # KeyError: storage is not yet initialized for `id`
        _registered_methods['on_tick'][mid] = {}
        _registered_methods['on_tick'][mid][timer_id] = f



"""Initialize modules."""
from .Heartbeat import Heartbeat
from .BaseHealthCheck import BaseHealthCheck

def init(config, uuclient):

    def cmd_heartbeat(dto_in):
        uucmd = config["uuApp"]['uuCmdList']['gatewayHeartbeat']
        return uuclient.post_request(uucmd, dto_in)

    gateway_config = config["gateway"]

    return [Heartbeat(cmd_heartbeat)]




""" Communication with the \*.Main uuSubApp. """

import datetime
import http.client
import json
import logging
import time
from typing import Dict, List, Optional, Tuple, Union

import requests
from requests_toolbelt import user_agent
from requests_toolbelt.multipart.encoder import MultipartEncoder

import uun_iot

logger = logging.getLogger(__name__)

TOKEN_LEEWAY = 60  # [s] - leeway between Client and Server Clock


class CommandError(Exception):
    def __init__(self, status, uu_app_error_map):
        super(__class__, self).__init__(uu_app_error_map)
        self._status = status
        self._code = list(uu_app_error_map.keys())[0]
        self._message = uu_app_error_map[self._code]["message"]

    def __str__(self):
        return str(self._status) + "," + self._code + "," + self._message


class UuAppClient:
    """Library functions for communication and authentication with uuApp.

    The class gets a dictionary ``config`` containing the configuration of the application.
    The (root) keys and subkeys of interest to this class are

        - ``uuApp``: information about server uuApp

            - ``gateway``: domain gateway of the corresponding uuApp
            - ``uuAppName``: full name of the uuApp
            - ``awid``
            - ``uuCmdList``: containing uuCmd endpoint. The keys are not used
              by :class:`UuAppClient` class - they are used by the application
              user modules (most commonly specified in Python package path
              ``<package_name>/modules/__init__.py``.

        - ``uuThing``: authentication information about the IoT gateway

            - ``uuIdentity``
            - ``accessCode1``
            - ``accessCode2``

        - ``oidcGrantToken``: information about the authentication token
          service used in server communication

            - ``gateway``: (usually ``uuidentity.plus4u.net``) domain gateway of the `OIDC` service
            - ``uuAppName``: (usually ``uu-oidc-maing02``) full name of the `OIDC` uuApp
            - ``awid``
            - ``clientId``
            - ``clientSecret``
            - ``uuCmd``: (usually ``oidc/grantToken``) uuCmd enpoint for token granting
            - ``tokenPath``: (usually ``./oidc-token``) filesystem location (on
              the IoT gateway) where the token will be saved

    Args:
        config: dictionary with configuration, see above
    """

    _config: dict
    _token: Optional[str]

    def __init__(self, config: dict):
        self._config = config
        self._token = None
        self._token_expires_at = None
        self._user_agent = user_agent("uun-iot", uun_iot.__version__)

        if (
            "uuApp" in self._config
            and "uuThing" in self._config
            and "oidcGrantToken" in self._config
        ):
            self._refresh_token()
        else:
            logger.info(
                "UuApp connection & gateway authentication is not "
                "defined in the configuration file. Are you sure, you do not "
                "want to use the connection? It will not be possible to "
                "issue requests to uu servers."
            )
            logger.debug("Not granting any auth token.")

    def _store_token(self, token_json: dict) -> None:
        with open(self._config["oidcGrantToken"]["tokenPath"], "w") as file:
            file.write(json.dumps(token_json))

    def _refresh_token(self) -> bool:
        """Returns False if token refresh was not needed - iff token is
        already loaded. Refresh token and return True otherwise."""

        if self._token is None:
            # load from file
            token_path = self._config["oidcGrantToken"]["tokenPath"]
            try:
                with open(token_path) as json_file:
                    token = json.load(json_file)
                    self._token = token["id_token"]
                    self._token_expires_at = token["expires_at"]

                logger.debug(
                    "loaded token from %s, expires at [%s]",
                    token_path,
                    datetime.datetime.fromtimestamp(self._token_expires_at),
                )
            except IOError:
                logger.debug("loading token from %s failed, file not found", token_path)

        current_time = TOKEN_LEEWAY + int(time.time())

        if self._token_expires_at is None or self._token_expires_at < current_time:
            # token expired
            self._grant_token()
            logger.debug(
                "granted new token, expires at [%s]",
                datetime.datetime.fromtimestamp(self._token_expires_at),
            )
            return True

        return False

    def _grant_token(self) -> None:
        """Grant token.

        Refresh and save the new token.

        Raises:
            CommandError: [TODO:description]
        """
        scope_host = self._config["uuApp"]["gateway"]
        scope_context = self._config["uuApp"]["uuAppName"]
        scope_awid = self._config["uuApp"]["awid"]

        post_data = {
            "grant_type": "password",
            "username": None,
            "password": None,
            "accessCode1": self._config["uuThing"]["accessCode1"],
            "accessCode2": self._config["uuThing"]["accessCode2"],
            "scope": f"openid https://{scope_host}/{scope_context}/{scope_awid}",
        }

        # does not matter if these are not present - not required - depends on server settings
        try:
            post_data.update(
                {
                    "client_id": self._config["oidcGrantToken"]["clientId"],
                    "client_secret": self._config["oidcGrantToken"]["clientSecret"],
                }
            )
        except KeyError:
            pass

        headers = {"Content-Type": "application/json", "User-Agent": self._user_agent}

        gateway = self._config["oidcGrantToken"]["gateway"]
        uu_app_name = self._config["oidcGrantToken"]["uuAppName"]
        awid = self._config["oidcGrantToken"]["awid"]
        uucmd = self._config["oidcGrantToken"]["uuCmd"]

        conn = http.client.HTTPSConnection(gateway)
        uri = "/%s/%s/%s" % (uu_app_name, awid, uucmd)
        conn.request("POST", uri, json.dumps(post_data), headers)
        response = conn.getresponse()

        if response.status >= 200 and response.status < 300:
            token_json = json.loads(response.read())
            self._token = token_json["id_token"]

            token_json["expires_at"] = int(time.time()) + int(token_json["expires_in"])
            self._token_expires_at = token_json["expires_at"]
            self._store_token(token_json)
        else:
            status = response.status
            error_json = json.loads(response.read())
            raise CommandError(status, error_json["uuAppErrorMap"])

    def _get_full_url(self, uucmd: str) -> str:
        gateway = self._config["uuApp"]["gateway"]
        uu_app_name = self._config["uuApp"]["uuAppName"]
        awid = self._config["uuApp"]["awid"]

        uri = "/%s/%s/%s" % (uu_app_name, awid, uucmd)
        return "https://%s%s" % (gateway, uri)

    def get_request(
        self,
        uucmd: str,
        dto_in: dict = {},
        http_error_level: int | None = logging.WARNING,
    ) -> requests.Response:
        """GET request with oidc2 authentication.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be passed as JSON input to the uuApp
            http_error_level: logging library level of HTTP error (HTTP status
                code not in 200 category), default logging.WARNING
        """
        self._refresh_token()

        headers = {
            "Authorization": "Bearer " + self._token,
            "Content-type": "application/json",
            "User-Agent": self._user_agent,
        }

        full_url = self._get_full_url(uucmd)

        logger.debug("`%s` GET request, payload: %s", uucmd, dto_in)
        response = requests.get(full_url, headers=headers, json=dto_in, timeout=20)

        if http_error_level is not None:
            if not 200 <= response.status_code < 300 and http_error_level is not None:
                logger.log(
                    http_error_level,
                    "`%s` GET response (%s): %s",
                    uucmd,
                    response.status_code,
                    response.text,
                )

        return response

    def get(
        self, uucmd: str, dto_in: dict = {}, log_level: int = logging.WARNING
    ) -> Tuple[requests.Response | None, Tuple[bool, requests.RequestException | None]]:
        """Get request

        Get request using a Bearer token. Connection errors are suppressed, logged and returned.

        Args:
            uucmd: UuCmd in relative path format, eg. "gateway/heartbeat"
            dto_in: data to send
            log_level: optional logging level for possible exceptions

        Returns:
            Tuple ``(response, (ok, exc))``,
                :class:`requests.Response` response, ``None`` if exception was raised
                boolean ``ok`` is True if no exception was encountered, False otherwise
                :class:`requests.RequestException` exc is None if no exception
                    during request was raised
        """

        response = None
        ok = False
        exc = None
        try:
            response = self.get_request(uucmd, dto_in, http_error_level=None)
            response.raise_for_status()
            ok = True
        except requests.HTTPError as e:
            logger.log(
                log_level,
                "`%s` GET response (%s): %s",
                uucmd,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(log_level, "`%s` GET RequestException: %s", uucmd, str(e))
            exc = e

        return response, (ok, exc)

    def post_request(
        self,
        uucmd: str,
        dto_in: dict = {},
        http_error_level: int | None = logging.WARNING,
    ) -> requests.Response:
        """POST request with oidc2 authentication.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be passed as JSON input to the uuApp
            http_error_level: logging library level of HTTP error (HTTP status
                code not in 200 category), default logging.WARNING
        """
        self._refresh_token()

        headers = {
            "Authorization": "Bearer " + self._token,
            "Content-type": "application/json",
            "User-Agent": self._user_agent,
        }

        full_url = self._get_full_url(uucmd)

        logger.debug("`%s` POST request, payload: %s", uucmd, dto_in)
        response = requests.post(full_url, headers=headers, json=dto_in)

        if http_error_level is not None:
            if not 200 <= response.status_code < 300 and http_error_level is not None:
                logger.log(
                    http_error_level,
                    "`%s` POST response (%s): %s",
                    uucmd,
                    response.status_code,
                    response.text,
                )

        return response

    def post(
        self, uucmd: str, dto_in: dict = {}, log_level: int = logging.WARNING
    ) -> Tuple[requests.Response | None, Tuple[bool, requests.RequestException | None]]:
        """Post request

        Post request using a Bearer token. Connection errors are suppressed, logged and returned.

        Args:
            uucmd: UuCmd in relative path format, eg. "gateway/heartbeat"
            dto_in: data to send
            log_level: optional logging level for possible exceptions

        Returns:
            Tuple ``(response, (ok, exc))``,
                :class:`requests.Response` response, ``None`` if exception was raised
                boolean ``ok`` is True if no exception was encountered, False otherwise
                :class:`requests.RequestException` exc is None if no exception
                    during request was raised
        """

        response = None
        ok = False
        exc = None
        try:
            response = self.post_request(uucmd, dto_in, http_error_level=None)
            response.raise_for_status()
            ok = True
        except requests.HTTPError as e:
            logger.log(
                log_level,
                "`%s` POST response (%s): %s",
                uucmd,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(log_level, "`%s` POST RequestException: %s", uucmd, str(e))
            exc = e

        return response, (ok, exc)

    def multipart_request(
        self, uucmd: str, dto_in: dict, http_error_level: int | None = logging.WARNING
    ):
        """POST request with oidc2 authentication and MULTIPART encoded data
        with oidc2 authentication. Useful for sending binary data (images, ...). See
        https://toolbelt.readthedocs.io/en/latest/user.html#multipart-form-data-encoder
        for information about multipart encoder.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be encoded in streaming multipart form-data object to the uuApp
            http_error_level: logging library level of HTTP error (HTTP status
                code not in 200 category), default logging.WARNING
        """
        self._refresh_token()

        full_url = self._get_full_url(uucmd)

        multipart_encoded = MultipartEncoder(fields=dto_in)
        headers = {
            "Authorization": "Bearer " + self._token,
            "Content-Type": multipart_encoded.content_type,
            "User-Agent": self._user_agent,
        }
        response = requests.post(full_url, headers=headers, data=multipart_encoded)

        if http_error_level is not None:
            if not 200 <= response.status_code < 300 and http_error_level is not None:
                logger.log(
                    http_error_level,
                    "`%s` MULTIPART response (%s): %s",
                    uucmd,
                    response.status_code,
                    response.text,
                )

        return response

    def multipart(
        self, uucmd: str, dto_in: dict = {}, log_level: int = logging.WARNING
    ) -> Tuple[requests.Response | None, Tuple[bool, requests.RequestException | None]]:
        """Multipart request

        Multipart request using a Bearer token. Connection errors are suppressed, logged and returned.

        Args:
            uucmd: UuCmd in relative path format, eg. "gateway/heartbeat"
            dto_in: data to send
            log_level: optional logging level for possible exceptions

        Returns:
            Tuple ``(response, (ok, exc))``,
                :class:`requests.Response` response, ``None`` if exception was raised
                boolean ``ok`` is True if no exception was encountered, False otherwise
                :class:`requests.RequestException` exc is None if no exception
                    during request was raised
        """

        response = None
        ok = False
        exc = None
        try:
            response = self.multipart_request(uucmd, dto_in, http_error_level=None)
            response.raise_for_status()
            ok = True
        except requests.HTTPError as e:
            logger.log(
                log_level,
                "`%s` POST response (%s): %s",
                uucmd,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(log_level, "`%s` POST RequestException: %s", uucmd, str(e))
            exc = e

        return response, (ok, exc)


class UuCmdSession:
    """Send all data with UuCmd in one session to prevent multiple connections."""

    def __init__(self, uuclient, uucmd, http_error_level=logging.WARNING):
        self._uucmd = uucmd
        self._loglevel = http_error_level

        uuclient._refresh_token()
        self._url = uuclient._get_full_url(uucmd)

        self._session = requests.Session()
        self._session.headers.update(
            {
                "Authorization": "Bearer " + uuclient._token,
                "Content-type": "application/json",
                "User-Agent": uuclient._user_agent,
            }
        )

    def get(
        self, data
    ) -> Tuple[requests.Response | None, Tuple[bool, requests.RequestException | None]]:
        """Get request."""
        logger.debug("`%s` GET request, payload: %s", self._uucmd, data)

        response = None
        ok = False
        exc = None
        try:
            response = self._session.post(self._url, json=data)
            response.raise_for_status()
            ok = True
        except requests.HTTPError as e:
            logger.log(
                self._loglevel,
                "`%s` POST response (%s): %s",
                self._uucmd,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(
                self._loglevel, "`%s` POST RequestException: %s", self._uucmd, str(e)
            )
            exc = e

        return response, (ok, exc)

    def get_raw(self, data) -> Tuple[requests.Response, bool]:
        """Get request."""
        logger.debug("`%s` GET request, payload: %s", self._uucmd, data)

        response = self._session.get(self._url, json=data)

        if not 200 <= response.status_code < 300:
            logger.log(
                self._loglevel,
                "`%s` GET response (%s): %s",
                self._uucmd,
                response.status_code,
                response.text,
            )
            return response, False

        return response, True

    def post(
        self, data
    ) -> Tuple[requests.Response | None, Tuple[bool, requests.RequestException | None]]:
        """Post request."""
        logger.debug("`%s` POST request, payload: %s", self._uucmd, data)

        response = None
        ok = False
        exc = None
        try:
            response = self._session.post(self._url, json=data)
            response.raise_for_status()
            ok = True
        except requests.HTTPError as e:
            logger.log(
                self._loglevel,
                "`%s` POST response (%s): %s",
                self._uucmd,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(
                self._loglevel, "`%s` POST RequestException: %s", self._uucmd, str(e)
            )
            exc = e

        return response, (ok, exc)

    def post_raw(self, data) -> Tuple[requests.Response, bool]:
        logger.debug("`%s` POST, payload: %s", self._uucmd, data)

        response = self._session.post(self._url, json=data)

        if not 200 <= response.status_code < 300:
            logger.log(
                self._loglevel,
                "`%s` GET response (%s): %s",
                self._uucmd,
                response.status_code,
                response.text,
            )
            return response, False

        return response, True


#
#    def post_ok(self, data) -> Tuple[requests.Response|requests.RequestException|None, bool]
#
#    def post(self, data) -> Tuple[requests.Response|requests.RequestException|None, bool]:
#        logger.debug("`%s` POST, payload: %s", self._uucmd, data)
#
#        ok = False
#        try:
#            response = self._session.post(self._url, json=data)
#            response.raise_for_status()
#            ok=True
#        except requests.HTTPError:
#            logger.log(self._loglevel, "`%s` POST response (%s): %s", self._uucmd, response.status_code, response.text)
#        except requests.RequestException as e:
#            response = None
#            logger.log(self._loglevel, "`%s` POST RequestException: %s", self._uucmd, str(e))
#
#        return response, ok
#



from abc import abstractmethod, ABC
from enum import Enum, auto
import logging
from typing import Callable, List, Optional, Any, Union

from uun_iot.utils import module_id, Storage
from uun_iot.typing import IModule

logger = logging.getLogger(__name__)

class ConfigScopeEnum(Enum):
    """ Enumeration for configuration scopes. """

    #: Use the root ``gateway`` key for the module configuration.
    GATEWAY=auto()

    #: Use the ``gateway.moduleId`` key for the module configuration.
    SELF=auto()

class Module(IModule, ABC):
    """ Provides utility functions for modules.

    Class :class:`Module` implements interface :class:`uun_iot.typing.IModule`
    for derived classes automatically. The :attr:`id` attribute is derived from
    the class name using :func:`uun_iot.utils.module_id_from_str`. If the class
    already has the :attr:`id` attribute, keep it.

    The derived classes must call the :class:`Module`'s init function using :func:`super`. Example:

    .. code-block:: python

        from uun_iot import Module
        class TestModule(Module):
            def __init__(self, config, uucmd):
                super().__init__(config=config, uucmd=uucmd)
                self.attribute1 = "value1"

    The utility functions include:

        - configuration management
        - storage management
        - backup and restore of storage to filesystem (useful for power failures)
        - automation for sending storage to uuApp
        - handling of failed send storage actions (useful for unstable internet connection)

    Configuration management:

        Access configuration in easy way as ``self._c("key/subkey/subsubkey")``
        instead of cumbersome ``self._config["key"]["subkey"]["subsubkey"]`` and more.
        See configuration manager :meth:`._config_manager` or its alias :meth:`._c`.

    Storage management:

        An instanced thread-safe :class:`~uun_iot.utils.Storage` is available as
        :attr:`._storage`. The class provides backup and restore to
        JSON files and limiting the storage in number of its entries. If
        enabled, oldest entries will be removed for the sake of newer ones.

    If the derived class does not have the ``_config`` attribute, set it to the
    constructor argument ``config``.

    Args:
        uucmd: function(storage)->failed_send_data to send data to uuApp server
        config: gateway configuration
    """
    #: is set in :meth:`__init__` dynamically according to derived class
    id: str = None

    #: optional uuCmd function for data sending
    _uucmd: Optional[Callable[..., List]]
    #: gateway configuration
    _config: dict

    #: storage facility
    _storage: Storage
    #: is backup to file active?
    __backup: bool

    def __init__(self, config: dict={}, uucmd: Callable[[Any], List]=None):

        self._uucmd = uucmd

        if not hasattr(self, "id") or (hasattr(self, "id") and self.id is None):
            self.id = module_id(self.__class__)

        if not hasattr(self, "_config"):
            self._config = config

        # load backup file with previous data
        # if module is supposed to be backed up
        if "moduleBackupStorage" in self._config and self.id in self._config["moduleBackupStorage"]:
            try:
                backup_path = self._config["moduleBackupStorage"][self.id]["path"]
                storage_limit = self._config["moduleBackupStorage"][self.id]["limit"]
                if not isinstance(storage_limit, int):
                    logger.warning("Configuration for `%s`: "
                    "limit of 'gateway.moduleBackupStorage' must be an integer.",
                    self.id)
                    raise TypeError
            except (KeyError, TypeError):
                backup_path = self._config["moduleBackupStorage"][self.id]
                storage_limit = 0
        else:
            backup_path = None
            storage_limit = 0

        self.__backup = backup_path is not None
        self._storage = Storage(self.id, backup_path, storage_limit)

    def _config_manager(self, key: Optional[str]=None, scope: ConfigScopeEnum=ConfigScopeEnum.SELF) -> Any:
        """ Return configuration key.

        Examples:
            
            Following example JSON configuration file with Python outputs:

            .. code-block:: json
                
                {
                    "gateway": {
                        "testModule1": {
                            "option1": "value1",
                            "structure": {
                                "field1": 1,
                                "field2": {
                                    "nested": 2
                                }
                            }
                        },

                        "testModule2": {
                            "option2": "value2"
                        }
                    }

            .. code-block:: python

                from uun_iot import ConfigScopeEnum

                # inside of a method of class TestModule1 derived from Module
                print(self._c("option1"))
                >>> value1

                print(self._c("structure/field1/nested"))
                >>> 2

                print(self._c("option1", scope=ConfigScopeEnum.SELF))
                >>> value1

                print(self._c("testModule1/option1", scope=ConfigScopeEnum.GATEWAY))
                >>> value1

                print(self._c("testModule2/option2", scope=ConfigScopeEnum.GATEWAY))
                >>> value2

        Args:
            key: specify value of configuration key to read. Set to ``None`` to
                return whole dictionary of corresponding configuration given by
                ``scope``. delimit subdictionaries with ``/`` symbol, ie.
                ``self._c("color/bright/red")`` is the same as
                ``self._c("color")["bright"]["red"]``

            scope: the root of configuration ``scope`` can be any of
                :class:`ConfigScopeEnum`. Defaults to :attr:`ConfigScopeEnum.SELF`,
                ie. ``key`` is taken as a subkey of ``gateway.<moduleId>``.
                Alternatively, specify :attr:`ConfigScopeEnum.GATEWAY` to take
                ``key`` as subkey of ``gateway`` key

        Returns:
            Any: the configuration entry

        Raises:
            KeyError: when a given key does not exist inside the configuration
        """

        mconf = self._config

        if key is None:
            return mconf

        if scope == ConfigScopeEnum.SELF:
            key = self.id + "/" + key

        key_parts = key.split("/")
        for k in key_parts:
            try:
                mconf = mconf[k]
            except KeyError:
                raise KeyError(f"key `{key}` is not present in configuration")

        return mconf

    def _send_data(self, data: Union[List, Any], usecase: Callable=None):
        """Logging wrapper for sending ``data`` with ``usecase`` function.

        Log a message with INFO level with information about the data, then send the data.

        Args:
            data: data to be sent by usecase as ``usecase(data)``
            usecase: if ``None``, use stored :attr:`._uucmd`

        Raises:
            TypeError: if ``usecase`` is not specified and ``self._uucmd`` is also not specified
        """
        if usecase is None:
            usecase = self._uucmd
        if usecase is None:
            raise TypeError(f"No usecase was specified and the module `{self.id}` "
                "has no stored uucmd." )

        try:
            length = len(data)
            logger.info("Sending data object for `%s` with [%i] entries", self.id, length)
        except TypeError:
            logger.info("Sending data object for `%s`", self.id)

        usecase(data)

    def _send_storage(self, usecase: Callable[[List], List]=None):
        """ Sends entire storage of module to uuApp via ``usecase`` or stored :attr:`uucmd`.

        The ``usecase`` takes data and outputs items which failed to be sent.
        This method saves these failed attempts to file and retries again
        together with new data on next envocation. Only failed entries will be
        saved to the backup file. This functionality is independent on the
        backup functionality.

        Args:
            usecase: function ``List->List`` with one argument taking the data
            to be sent and outputting list with items which were not sent. If
            ``None``, use ``self._uucmd``

        Raises:
            TypeError: if ``usecase`` is not specified and ``self._uucmd`` is also not specified
        """

        if self._storage.is_empty():
            # no data to send
            return

        if usecase is None:
            usecase = self._uucmd
        if usecase is None:
            raise TypeError(f"No usecase was specified and the module `{self.id}` "
                "has no stored uucmd." )

        # prevent situation, where data are sent, another thread writes data to
        # storage, and (as a consequence of sending data) storage is emptied
        # thus discarding the latest unsent data
        with self._storage.write_lock:
            length = len(self._storage)
            logger.info("Sending storage of `%s` with [%i] entries", self.id, length)

            failed_entries = usecase(self._storage.data)
            self._storage.empty()

        if not self.__backup:
            # do not backup
            return

        if len(failed_entries) == 0:
            # everything was sent
            logger.info("Whole storage of `%s` sent successfully", self.id)
        else:
            # merge the two objects (chronologically) to prevent discarding
            # entries added to ._storage while previus ones were being
            # processed/sent

            # extend the storage with entries of the failed_entries list
            self._storage.merge(failed_entries, new_data=False)
            logger.warning("Could not send %i entries from storage of `%s`",
                    len(failed_entries), self.id)

    _c = _config_manager
